from __future__ import annotations

import os
import pickle
import shutil
from pathlib import Path
from typing import List

import cv2
import numpy as np
from tqdm import tqdm

from supervisely.io.json import load_json_file
from supervisely.nn.benchmark.base_evaluator import BaseEvalResult, BaseEvaluator
from supervisely.nn.benchmark.semantic_segmentation.metric_provider import (
    MetricProvider,
)
from supervisely.nn.benchmark.utils import (
    calculate_semsegm_metrics as calculate_metrics,
)
from supervisely.project.project import Dataset, OpenMode, Project
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger


class SemanticSegmentationEvalResult(BaseEvalResult):
    mp_cls = MetricProvider

    def _read_files(self, path: str) -> None:
        """Read all necessary files from the directory"""

        eval_data_path = Path(path) / "eval_data.pkl"
        if eval_data_path.exists():
            with open(Path(path, "eval_data.pkl"), "rb") as f:
                self.eval_data = pickle.load(f)

        inference_info_path = Path(path) / "inference_info.json"
        if inference_info_path.exists():
            self.inference_info = load_json_file(str(inference_info_path))

        speedtest_info_path = Path(path).parent / "speedtest" / "speedtest.json"
        if speedtest_info_path.exists():
            self.speedtest_info = load_json_file(str(speedtest_info_path))

    def _prepare_data(self) -> None:
        """Prepare data to allow easy access to the most important parts"""

        self.mp = MetricProvider(self.eval_data)

    @classmethod
    def from_evaluator(
        cls, evaulator: SemanticSegmentationEvaluator
    ) -> SemanticSegmentationEvalResult:
        """Method to customize loading of the evaluation result."""
        eval_result = cls()
        eval_result.eval_data = evaulator.eval_data
        eval_result._prepare_data()
        return eval_result

    @property
    def key_metrics(self):
        return self.mp.json_metrics()


class SemanticSegmentationEvaluator(BaseEvaluator):
    EVALUATION_PARAMS_YAML_PATH = f"{Path(__file__).parent}/evaluation_params.yaml"
    eval_result_cls = SemanticSegmentationEvalResult

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bg_cls_name = None
        self.bg_cls_color = None

    def evaluate(self):
        self.bg_cls_name = self._get_bg_class_name()
        if self.bg_cls_name not in self.classes_whitelist:
            self.classes_whitelist.append(self.bg_cls_name)

        gt_prep_path, pred_prep_path = self.prepare_segmentation_data()

        self.eval_data = calculate_metrics(
            gt_dir=gt_prep_path,
            pred_dir=pred_prep_path,
            boundary_width=0.01,
            boundary_iou_d=0.02,
            num_workers=4,
            class_names=self.classes_whitelist,
            result_dir=self.result_dir,
            progress=self.pbar,
        )
        self.eval_data["bg_cls_name"] = self.bg_cls_name
        logger.info("Successfully calculated evaluation metrics")
        self._dump_eval_results()
        logger.info("Evaluation results are saved")

    def _get_palette(self, project_path):
        meta_path = Path(project_path) / "meta.json"
        meta = ProjectMeta.from_json(load_json_file(meta_path))

        palette = [obj.color for obj in meta.obj_classes if obj.name in self.classes_whitelist]

        return palette

    def _dump_eval_results(self):
        eval_data_path = self._get_eval_data_path()
        self._dump_pickle(self.eval_data, eval_data_path)  # TODO: maybe dump JSON?

    def _get_eval_data_path(self):
        base_dir = self.result_dir
        eval_data_path = os.path.join(base_dir, "eval_data.pkl")
        return eval_data_path

    def prepare_segmentation_data(self):
        src_dirs = [self.gt_project_path, self.pred_project_path]
        output_dirs = [
            Path(self.gt_project_path).parent / "preprocessed_gt",
            Path(self.pred_project_path).parent / "preprocessed_pred",
        ]

        for src_dir, output_dir in zip(src_dirs, output_dirs):
            if output_dir.exists():
                logger.info(f"Preprocessed data already exists in {output_dir} directory")
                continue

            palette = self._get_palette(src_dir)
            bg_color = palette[self.classes_whitelist.index(self.bg_cls_name)]
            output_dir.mkdir(parents=True)
            temp_seg_dir = src_dir + "_temp"
            if not os.path.exists(temp_seg_dir):
                Project.to_segmentation_task(
                    src_dir,
                    temp_seg_dir,
                    target_classes=self.classes_whitelist,
                    bg_name=self.bg_cls_name,
                    bg_color=bg_color,
                )

            palette_lookup = np.zeros(256**3, dtype=np.int32)
            for idx, color in enumerate(palette, 1):
                key = (color[0] << 16) | (color[1] << 8) | color[2]
                palette_lookup[key] = idx

            temp_project = Project(temp_seg_dir, mode=OpenMode.READ)
            temp_project.total_items
            for dataset in temp_project.datasets:
                dataset: Dataset
                names = dataset.get_items_names()
                for name in names:
                    mask_path = dataset.get_seg_path(name)
                    mask = cv2.imread(mask_path)[:, :, ::-1]

                    mask_keys = (
                        (mask[:, :, 0].astype(np.int32) << 16)
                        | (mask[:, :, 1].astype(np.int32) << 8)
                        | mask[:, :, 2].astype(np.int32)
                    )
                    result = palette_lookup[mask_keys]
                    if name.count(".png") > 1:
                        name = name[:-4]
                    cv2.imwrite(os.path.join(output_dir, name), result)

            shutil.rmtree(temp_seg_dir)

        return output_dirs

    def _get_bg_class_name(self):
        possible_names = ["background", "bg", "unlabeled", "neutral", "__bg__"]
        logger.info(f"Searching for background class in projects. Possible names: {possible_names}")

        bg_cls_names = []
        for project_path in [self.gt_project_path, self.pred_project_path]:
            meta_path = Path(project_path) / "meta.json"
            meta = ProjectMeta.from_json(load_json_file(meta_path))

            for obj_cls in meta.obj_classes:
                if obj_cls.name in possible_names:
                    bg_cls_names.append(obj_cls.name)
                    break

        if len(bg_cls_names) == 0:
            raise ValueError("Background class not found in GT and Pred projects")

        if len(set(bg_cls_names)) > 1:
            raise ValueError(
                f"Founds multiple background class names in GT and Pred projects: {set(bg_cls_names)}"
            )

        return bg_cls_names[0]



# isort: skip_file
from supervisely.convert.base_converter import (
    AvailableImageConverters,
    AvailablePointcloudConverters,
    AvailableVideoConverters,
    AvailableVolumeConverters,
    BaseConverter,
)
from supervisely.convert.converter import ImportManager

# Image
from supervisely.convert.image.cityscapes.cityscapes_converter import (
    CityscapesConverter,
)
from supervisely.convert.image.coco.coco_converter import COCOConverter
from supervisely.convert.image.coco.coco_anntotation_converter import FastCOCOConverter
from supervisely.convert.image.csv.csv_converter import CSVConverter
from supervisely.convert.image.multispectral.multispectral_converter import (
    MultiSpectralImageConverter,
)
from supervisely.convert.image.medical2d.medical2d_converter import Medical2DImageConverter
from supervisely.convert.image.masks.images_with_masks_converter import ImagesWithMasksConverter

from supervisely.convert.image.pascal_voc.pascal_voc_converter import PascalVOCConverter
from supervisely.convert.image.pdf.pdf_converter import PDFConverter
from supervisely.convert.image.sly.sly_image_converter import SLYImageConverter
from supervisely.convert.image.sly.fast_sly_image_converter import FastSlyImageConverter
from supervisely.convert.image.yolo.yolo_converter import YOLOConverter
from supervisely.convert.image.multi_view.multi_view import MultiViewImageConverter
from supervisely.convert.image.label_me.label_me_converter import LabelmeConverter
from supervisely.convert.image.label_studio.label_studio_converter import LabelStudioConverter
from supervisely.convert.image.high_color.high_color_depth import HighColorDepthImageConverter


# Pointcloud
from supervisely.convert.pointcloud.sly.sly_pointcloud_converter import SLYPointcloudConverter
from supervisely.convert.pointcloud.las.las_converter import LasConverter
from supervisely.convert.pointcloud.ply.ply_converter import PlyConverter
from supervisely.convert.pointcloud.bag.bag_converter import BagConverter


# Pointcloud Episodes
from supervisely.convert.pointcloud_episodes.sly.sly_pointcloud_episodes_converter import (
    SLYPointcloudEpisodesConverter,
)
from supervisely.convert.pointcloud_episodes.bag.bag_converter import BagEpisodesConverter

# Video
from supervisely.convert.video.mot.mot_converter import MOTConverter
from supervisely.convert.video.sly.sly_video_converter import SLYVideoConverter

# Volume
from supervisely.convert.volume.sly.sly_volume_converter import SLYVolumeConverter
from supervisely.convert.volume.dicom.dicom_converter import DICOMConverter



import os

from supervisely import (
    Annotation,
    Api,
    ProjectMeta,
    Label,
    Rectangle,
    batched,
    is_development,
    logger,
)
from supervisely.convert.image.sly.sly_image_converter import SLYImageConverter
import supervisely.convert.image.sly.sly_image_helper as helper
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.io.fs import get_file_ext
from supervisely.io.json import load_json_file
from supervisely.convert.image.image_helper import validate_image_bounds


class FastSlyImageConverter(SLYImageConverter, ImageConverter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._supports_links = False

    def validate_format(self) -> bool:

        detected_ann_cnt = 0
        self._items = []
        meta = ProjectMeta()
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if ext == self.ann_ext:
                    ann_json = load_json_file(full_path)
                    if helper.annotation_high_level_validator(ann_json):
                        meta = helper.get_meta_from_annotation(ann_json, meta)
                        h, w = helper.get_image_size_from_annotation(ann_json)
                        image_name = os.path.splitext(os.path.basename(full_path))[0]
                        item = self.Item(image_name)
                        item.ann_data = full_path
                        item.set_shape((h, w))
                        self._items.append(item)
                        detected_ann_cnt += 1
                elif self.is_image(full_path):
                    self._items = []
                    return False

        if detected_ann_cnt == 0:
            self._items = []
            return False
        self._meta = meta
        return detected_ann_cnt > 0

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        """Convert to Supervisely format."""
        if meta is None:
            meta = self._meta

        try:
            ann_json = load_json_file(item.ann_data)
            if "annotation" in ann_json:
                ann_json = ann_json["annotation"]
            if renamed_classes or renamed_tags:
                ann_json = helper.rename_in_json(ann_json, renamed_classes, renamed_tags)
            img_size = list(ann_json["size"].values())
            labels = validate_image_bounds(
                [Label.from_json(obj, meta) for obj in ann_json["objects"]],
                Rectangle.from_size(img_size),
            )
            return Annotation.from_json(ann_json, meta).clone(labels=labels)
        except Exception as e:
            logger.warn(f"Failed to convert annotation: {repr(e)}")
            return None

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 50,
        log_progress=True,
    ) -> None:
        """Convert Supervisely annootations to Supervisely and append to dataset images."""

        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        existing_images = {img_info.name: img_info for img_info in api.image.get_list(dataset_id)}
        if len(existing_images) == 0:
            raise RuntimeError(
                "Not found images in the dataset. "
                "Please start the import process from a dataset that contains images, "
                "or upload both images and annotations at once."
            )
        if log_progress:
            progress, progress_cb = self.get_progress(
                self.items_count, "Adding Supervisely annotations..."
            )
        else:
            progress_cb = None

        for batch in batched(self._items, batch_size=batch_size):
            img_ids = []
            anns = []
            for item in batch:
                existing_image = existing_images.get(item.name)
                if existing_image is None:
                    continue
                if item.shape != (existing_image.height, existing_image.width):
                    logger.warn(
                        f"Image '{item.name}' has different shapes in JSON file and server."
                    )
                    continue

                ann = self.to_supervisely(item, meta, renamed_classes, renamed_tags)
                if ann is not None:
                    img_ids.append(existing_image.id)
                    anns.append(ann)

            if len(anns) == len(img_ids):
                existing_anns = api.annotation.download_json_batch(dataset_id, img_ids)
                existing_anns = [Annotation.from_json(json, meta) for json in existing_anns]
                merged_anns = []
                for existing_ann, new_ann in zip(existing_anns, anns):
                    merged_ann = existing_ann.merge(new_ann)
                    merged_anns.append(merged_ann)

                api.annotation.upload_anns(img_ids, merged_anns)

            if log_progress:
                progress_cb(len(batch))

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset ID:'{dataset_id}' has been successfully uploaded.")



# coding: utf-8
from __future__ import annotations

from typing import Dict, Optional, Union
from uuid import UUID

from supervisely._utils import take_with_default
from supervisely.geometry.geometry import Geometry
from supervisely.pointcloud_annotation.pointcloud_episode_object import (
    PointcloudEpisodeObject,
)
from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import (
    PointcloudEpisodeObjectCollection,
)
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject
from supervisely.pointcloud_annotation.pointcloud_object_collection import (
    PointcloudObjectCollection,
)
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_figure import VideoFigure


class PointcloudFigure(VideoFigure):
    """
    PointcloudFigure object for
    :class:`PointcloudAnnotation<supervisely.pointcloud_annotation.pointcloud_annotation.PointcloudAnnotation>` or :class:`PointcloudEpisodeAnnotation<supervisely.pointcloud_annotation.pointcloud_episode_annotation.PointcloudEpisodeAnnotation>`.
    :class:`PointcloudFigure<PointcloudFigure>` objects is immutable.

    :param parent_object: PointcloudObject or PointcloudObject object.
    :type parent_object: Union[PointcloudObject, PointcloudEpisodeObject]
    :param geometry: Label :class:`geometry<supervisely.geometry.geometry.Geometry>`.
    :type geometry: Geometry
    :param frame_index: Index of Frame to which PointcloudFigure belongs.
    :type frame_index: int
    :param key: KeyIdMap object.
    :type key: KeyIdMap, optional
    :param class_id: ID of :class:`PointcloudObject<PointcloudObject>` (or :class:`PointcloudEpisodeObject<PointcloudEpisodeObject>`) to which PointcloudFigure belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created PointcloudFigure.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when PointcloudFigure was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when PointcloudFigure was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

        obj_class_car = sly.ObjClass('car', Cuboid3d)
        pointcloud_obj_car = sly.PointcloudObject(obj_class_car)

        position, rotation, dimension = Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
        cuboid = Cuboid3d(position, rotation, dimension)
        frame_index = 10
        figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=frame_index)

        print(figure.to_json())
        # Output: {
        #     "geometry": {
        #         "dimensions": { "x": 1.8, "y": 3.9, "z": 1.6 },
        #         "position": { "x": -3.4, "y": 28.9, "z": -0.7 },
        #         "rotation": { "x": 0.0, "y": 0, "z": -0.03 }
        #     },
        #     "geometryType": "cuboid_3d",
        #     "key": "4beae1be12624b70ad533c8be7477605",
        #     "objectKey": "c1e1965efc0d4ae9b0b39367b04d637a"
        # }
    """

    def __init__(
        self,
        parent_object: Union[PointcloudObject, PointcloudEpisodeObject],
        geometry: Geometry,
        frame_index: Optional[int] = None,
        key: Optional[UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            parent_object,
            geometry,
            frame_index,
            key=key,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        # @TODO: validate geometry - allowed: only cuboid_3d + point_cloud

    @property
    def parent_object(self) -> Union[PointcloudObject, PointcloudEpisodeObject]:
        """
        PointcloudObject of current PointcloudFigure.

        :return: PointcloudObject ot PointcloudEpisodeObject object
        :rtype: :class:`PointcloudObject<PointcloudObject>` or :class:`PointcloudEpisodeObject<PointcloudEpisodeObject>`
        :Usage example:
         .. code-block:: python

            pointcloud_obj_car = pointcloud_figure_car.parent_object

            print(pointcloud_obj_car.to_json())
            # Output: {
            #     "key": "d573c6f081544e3da20022d932b259c1",
            #     "classTitle": "car",
            #     "tags": []
            # }
        """

        return super().parent_object

    @property
    def video_object(self) -> None:
        """Not supported for pointcloud."""

        raise NotImplementedError("If you faced this error, please write to technical support.")

    def validate_bounds(self, img_size, _auto_correct=False):
        """Not supported for pointcloud."""

        raise NotImplementedError()

    @classmethod
    def from_json(
        cls,
        data: Dict,
        objects: Union[PointcloudObjectCollection, PointcloudEpisodeObjectCollection],
        frame_index: Optional[int] = None,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> PointcloudFigure:
        """
        Convert a json dict to PointcloudFigure. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: dict
        :param objects: PointcloudObjectCollection or PointcloudEpisodeObjectCollection object.
        :type objects: PointcloudObjectCollection or PointcloudEpisodeObjectCollection
        :param frame_index: Index of Frame to which PointcloudFigure belongs.
        :type frame_index: int
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :raises: :class:`RuntimeError`, if point cloudobject ID and pointcloud object key are None, if pointcloud object key and key_id_map are None, if pointcloud object with given id not found in key_id_map
        :return: PointcloudFigure object
        :rtype: :class:`PointcloudFigure`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

            obj_class_car = sly.ObjClass('car', Cuboid3d)
            pointcloud_obj_car = sly.PointcloudObject(obj_class_car)

            position, rotation, dimension = Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
            cuboid = Cuboid3d(position, rotation, dimension)
            frame_index = 10
            figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=frame_index)
            pointcloud_figure_json = figure.to_json(save_meta=True)

            new_pointcloud_figure = sly.PointcloudFigure.from_json(
                pointcloud_figure_json,
                sly.PointcloudObjectCollection([pointcloud_obj_car]),
                frame_index
            )
        """

        return super().from_json(data, objects, frame_index, key_id_map)

    def clone(
        self,
        parent_object: Optional[Union[PointcloudObject, PointcloudEpisodeObject]] = None,
        geometry: Optional[Geometry] = None,
        frame_index: Optional[int] = None,
        key: Optional[UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> PointcloudFigure:
        """
        Makes a copy of PointcloudFigure with new fields, if fields are given, otherwise it will use fields of the original PointcloudFigure.

        :param parent_object: :class:`PointcloudObject<PointcloudObject>` (or :class:`PointcloudEpisodeObject<PointcloudEpisodeObject>`) object.
        :type parent_object: PointcloudObject or PointcloudEpisodeObject, optional
        :param geometry: Label :class:`geometry<supervisely.geometry.geometry.Geometry>`.
        :type geometry: Geometry, optional
        :param frame_index: Index of Frame to which PointcloudFigure belongs.
        :type frame_index: int, optional
        :param key: KeyIdMap object.
        :type key: KeyIdMap, optional
        :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which PointcloudFigure belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created PointcloudFigure.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when PointcloudFigure was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when PointcloudFigure was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :return: PointcloudFigure object
        :rtype: :class:`PointcloudFigure`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

            obj_class_car = sly.ObjClass('car', Cuboid3d)
            pointcloud_obj_car = sly.PointcloudObject(obj_class_car)

            position, rotation, dimension = Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
            cuboid = Cuboid3d(position, rotation, dimension)
            frame_index = 10
            figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=frame_index)

            # Remember that PointcloudFigure object is immutable, and we need to assign new instance of PointcloudFigure to a new variable
            pointcloud_figure_clone = figure.clone(parent_object=pointcloud_obj_car, frame_index=11)

            print(pointcloud_figure_clone.to_json())
            # Output: {
            #     "geometry": {
            #         "dimensions": { "x": 1.8, "y": 3.9, "z": 1.6 },
            #         "position": { "x": -3.4, "y": 28.9, "z": -0.7 },
            #         "rotation": { "x": 0.0, "y": 0, "z": -0.03 }
            #     },
            #     "geometryType": "cuboid_3d",
            #     "key": "4beae1be12624b70ad533c8be7477605",
            #     "objectKey": "c1e1965efc0d4ae9b0b39367b04d637a"
            # }
        """

        return self.__class__(
            parent_object=take_with_default(parent_object, self.parent_object),
            geometry=take_with_default(geometry, self.geometry),
            frame_index=take_with_default(frame_index, self.frame_index),
            key=take_with_default(key, self._key),
            class_id=take_with_default(class_id, self.class_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
        )
